/*
 * Access several filesystems.
 *
 * Copyright (C) 2018, 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/util/util.h>

#include <l4/re/c/dataspace.h>
#include <l4/re/env.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/err.h>
#include <l4/sys/kernel_object.h>
#include <l4/sys/types.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <fsclient/fs_client.h>
#include <fsclient/fsdesc.h>
#include <fsclient/fs_ipc.h>
#include <fsserver/server.h>
#include <ipc/util_ipc.h>
#include "fspath.h"



/* Virtual filesystem server. */

class Vfs_server : public OpeningServer
{
private:
  /* NOTE: Improvised structure! */

  l4_cap_idx_t _servers[10];
  const char *_mountpoints[10];
  int _mounted = 0;

public:
  explicit Vfs_server()
  : OpeningServer()
  {
  }

  void dispatch(ipc_message_t *msg)
  {
    switch (l4_msgtag_label(msg->tag))
    {
      case Fs_op_mount:
        mount(msg);
        break;

      case Fs_op_open:
        open(msg);
        break;

      case Fs_op_getfs:
        getfs(msg);
        break;

      default:
        send_error(msg, -L4_EBADPROTO);
        break;
    }
  }

  void getfs(ipc_message_t *msg)
  {
    _open(Fs_op_getfs, msg);
  }

  void open(ipc_message_t *msg)
  {
    _open(Fs_op_open, msg);
  }

  void _open(int op, ipc_message_t *msg)
  {
    int flags;
    l4_cap_idx_t server, irq;
    l4re_ds_t ds;
    size_t size;
    char *buffer;
    long err;

    /* Obtain the flags. */

    flags = ipc_message_get_word(msg, 0);

    /* Obtain the dataspace and IRQ object. */

    err = import_dataspace(msg, 0, &ds, (l4_addr_t *) &buffer) || import_capability(msg, 1, &irq);
    if (err)
    {
      send_error(msg, -L4_EIO);
      return;
    }

    /* Match the path to a mountpoint. */

    int selected = find_mountpoint(buffer);

    if (selected < 0)
    {
      send_error(msg, -L4_ENOENT);
      return;
    }

    /* Rewrite the path for the selected filesystem. */

    rewrite_path(buffer, _mountpoints[selected]);

    /* Find the path within the selected filesystem. */

    /* Propagate the message to the identified filesystem.
       NOTE: The L4_MSGTAG_PROPAGATE flag is not supported by Fiasco.OC, so this
       NOTE: object has to act as intermediary. */

    err = fs_ipc_open(flags, _servers[selected], op, ds, irq, &server, &size);
    if (err)
    {
      send_error(msg, -L4_EIO);
      return;
    }

    /* Free the capability and buffer from this task. */

    discard_dataspace(ds, (l4_addr_t) buffer);

    /* Return the file size. */

    ipc_message_add_word(msg, size);

    /* Export and eventually free the server reference from this task. */

    ipc_message_propagate_item(msg, server);
  }

  void mount(ipc_message_t *msg)
  {
    l4_cap_idx_t server;
    l4re_ds_t ds;
    char *buffer;
    long err;

    /* Obtain the filesystem capability. */

    err = import_capability(msg, 0, &server);
    if (err)
    {
      send_error(msg, -L4_EIO);
      return;
    }

    /* Obtain the dataspace. */

    err = import_dataspace(msg, 1, &ds, (l4_addr_t *) &buffer);
    if (err)
    {
      send_error(msg, -L4_EIO);
      return;
    }

    /* Obtain the object's path. */

    char *path = buffer;

    /* Mount the supplied capability. */

    mount_at_path(server, path);

    /* Free the capability and buffer from this task. */

    discard_dataspace(ds, (l4_addr_t) buffer);
  }

  /* Non-exported methods. */

  void mount_at_path(l4_cap_idx_t server, const char *path)
  {
    /* NOTE: Should return error, terminate with path separator if absent. */

    if (_mounted >= 10)
      return;

    _servers[_mounted] = server;
    _mountpoints[_mounted] = strdup(path);

    _mounted++;
  }

  int find_mountpoint(const char *path)
  {
    int i, matching;
    size_t length, longest;

    if (!_mounted)
      return -1;

    matching = -1;
    longest = 0;

    /* Find the largest matching path. */

    for (i = 0; i < _mounted; i++)
    {
      length = strlen(_mountpoints[i]);

      /* The empty mountpoint always matches;
         suffixed mountpoints may match as prefixes;
         suffixed mountpoints may match unsuffixed paths;
         unsuffixed mountpoints must match exactly. */

      /* NOTE: Unsuffixed mountpoints might support subpaths by insisting on
               a path separator as the first subsequent character. */

      if (!length ||
          ((_mountpoints[i][length - 1] == '/') && !strncmp(_mountpoints[i], path, length)) ||
          ((_mountpoints[i][length - 1] == '/') && (strlen(path) == length - 1) && !strncmp(_mountpoints[i], path, length - 1)) ||
          !strcmp(_mountpoints[i], path))
      {
        if (length >= longest)
          matching = i;
      }
    }

    return matching;
  }
};



int main(int argc, char *argv[])
{
  FILE *fp;
  char buffer[256], *sep;
  l4_cap_idx_t fscap, server;

  if (argc < 2)
  {
    printf("Need a mount table.\n");
    return 1;
  }

  /* Initialise and register a new server object. */

  Vfs_server server_obj;

  if (bind_to_capability("export", l4_umword_t(&server_obj), &server))
  {
    printf("Could not bind thread.\n");
    return 1;
  }

  /* Mount filesystems. */

  fp = fopen(argv[1], "r");
  if (fp == NULL)
  {
    printf("Could not open %s.\n", argv[1]);
    return 1;
  }

  /* Obtain capabilities for mountpoints. */

  while (fgets(buffer, 256, fp) != NULL)
  {
    sep = strchr(buffer, (int) '\t');
    if (sep == NULL)
      continue;

    /* Terminate newline-suffixed strings. */

    if (sep[strlen(sep) - 1] == '\n')
      sep[strlen(sep) - 1] = '\0';

    /* Obtain the named capability. */

    fscap = l4re_env_get_cap(sep + 1);

    if (l4_is_invalid_cap(fscap))
      continue;

    /* Terminate the path and register the mountpoint. */

    *sep = '\0';
    server_obj.mount_at_path(fscap, buffer);
  }

  fclose(fp);

  /* Enter the IPC server loop. */

  server_obj.loop();
  return 0;
}
